import decimal
import json
from datetime import date, datetime, time
from typing import Any, Dict, Optional

import dateutil.parser
from django.core.cache import cache
from django.core.files import File
from django.core.files.storage import default_storage
from django.db.models import Model

from hierarkey.models import Hierarkey


class HierarkeyProxy:
    """
    If you add a hierarkey storage to a model, the model will get a new attribute (e.g. ``settings``) containing
    a key-value store that is managed by this class.

    This class allows access to settings via attribute access, item access or using the documented methods.
    You should not instantiate this class yourself.
    """

    @classmethod
    def _new(cls, obj: Model, hierarkey: Hierarkey, cache_namespace: str, parent: Optional[Model] = None,
             type: type = None):
        o = HierarkeyProxy()
        o._obj = obj
        o._h = hierarkey
        o._cache_namespace = cache_namespace
        o._parent = parent
        o._cached_obj = None
        o._write_cached_obj = None
        o._type = type
        return o

    @property
    def _objects(self):
        """
        Returns a model manager (or related object manager) giving access to the raw objects backing this
        storage level.
        """
        return getattr(self._obj, '_%s_objects' % self._h.attribute_name)

    def _cache(self) -> Dict[str, Any]:
        if self._cached_obj is None:
            self._cached_obj = cache.get_or_set(
                'hierarkey_{}_{}'.format(self._cache_namespace, self._obj.pk),
                lambda: {s.key: s.value for s in self._objects.all()},
                timeout=1800
            )
        return self._cached_obj

    def _write_cache(self) -> Dict[str, Any]:
        if self._write_cached_obj is None:
            self._write_cached_obj = {
                s.key: s for s in self._objects.all()
            }
        return self._write_cached_obj

    def flush(self) -> None:
        """
        Discards both the state within this object as well as the cache in Django's cache backend.
        """
        self._cached_obj = None
        self._write_cached_obj = None
        self._flush_external_cache()

    def _flush_external_cache(self):
        cache.delete('hierarkey_{}_{}'.format(self._cache_namespace, self._obj.pk))

    def freeze(self) -> dict:
        """
        Returns a dictionary of all settings set for this object, including
        any values of its parents or hardcoded defaults.
        """
        settings = {}
        for key, v in self._h.defaults.items():
            settings[key] = self._unserialize(v.value, v.type)
        if self._parent:
            settings.update(getattr(self._parent, self._h.attribute_name).freeze())
        for key in self._cache():
            settings[key] = self.get(key)
        return settings

    def _unserialize(self, value: str, as_type: type, binary_file=False) -> Any:
        if as_type is None and value is not None and value.startswith('file://'):
            as_type = File

        if as_type is not None and isinstance(value, as_type):
            return value
        elif value is None:
            return None
        elif as_type == int or as_type == float or as_type == decimal.Decimal:
            return as_type(value)
        elif as_type == dict or as_type == list:
            return json.loads(value)
        elif as_type == bool or value in ('True', 'False'):
            return value == 'True'
        elif as_type == File:
            try:
                fi = default_storage.open(value[7:], 'rb' if binary_file else 'r')
                fi.url = default_storage.url(value[7:])
                return fi
            except OSError:
                return False
        elif as_type == datetime:
            return dateutil.parser.parse(value)
        elif as_type == date:
            return dateutil.parser.parse(value).date()
        elif as_type == time:
            return dateutil.parser.parse(value).time()
        elif as_type is not None:
            for t in self._h.types:
                if issubclass(as_type, t.type):
                    return t.unserialize(value)

        if as_type is not None and issubclass(as_type, Model):
            return as_type.objects.get(pk=value)

        return value

    def _serialize(self, value: Any) -> str:
        if isinstance(value, str):
            return value
        elif isinstance(value, int) or isinstance(value, float) \
                or isinstance(value, bool) or isinstance(value, decimal.Decimal):
            return str(value)
        elif isinstance(value, list) or isinstance(value, dict):
            return json.dumps(value)
        elif isinstance(value, datetime) or isinstance(value, date) or isinstance(value, time):
            return value.isoformat()
        elif isinstance(value, Model):
            return value.pk
        elif isinstance(value, File):
            return 'file://' + value.name
        else:
            for t in self._h.types:
                if isinstance(value, t.type):
                    return t.serialize(value)

        raise TypeError('Unable to serialize %s into a setting.' % str(type(value)))

    def get(self, key: str, default=None, as_type: type = None, binary_file: bool = False):
        """
        Get a setting specified by ``key``. Normally, settings are strings, but
        if you put non-strings into the settings object, you can request deserialization
        by specifying ``as_type``. If the key does not have a harcdoded default type,
        omitting ``as_type`` always will get you a string.

        If the setting with the specified name does not exist on this object, any parent object
        up to the global settings layer (if configured) will be queried. If still no value is
        found, a default value set in ths source code will be returned if one exists.
        If not, the value of the ``default`` argument of this method will be returned instead.

        If you receive a ``File`` object, it will already be opened. You can specify the ``binary_file``
        flag to indicate that it should be opened in binary mode.
        """
        if as_type is None:
            as_type = self._h.get_declared_type(key)

        if key in self._cache():
            value = self._cache()[key]
        else:
            value = None
            if self._parent:
                value = getattr(self._parent, self._h.attribute_name).get(key, as_type=str)
            if value is None and key in self._h.defaults:
                value = self._h.defaults[key].value
            if value is None and default is not None:
                value = default

        return self._unserialize(value, as_type, binary_file=binary_file)

    def __getitem__(self, key: str) -> Any:
        return self.get(key)

    def __getattr__(self, key: str) -> Any:
        if key.startswith('_'):  # pragma: no cover
            return super().__getattribute__(key)
        return self.get(key)

    def __setattr__(self, key: str, value: Any) -> None:
        if key.startswith('_'):
            return super().__setattr__(key, value)
        self.set(key, value)

    def __setitem__(self, key: str, value: Any) -> None:
        self.set(key, value)

    def set(self, key: str, value: Any) -> None:
        """
        Stores a setting in the database and connects it to its object.

        The write to the database is performed immediately and the cache in the cache backend is flushed.
        The cache within this object will be updated correctly.
        """
        wc = self._write_cache()
        if key in wc:
            s = wc[key]
        else:
            s = self._type(object=self._obj, key=key)
        s.value = self._serialize(value)
        s.save()
        self._cache()[key] = s.value
        wc[key] = s
        self._flush_external_cache()

    def __delattr__(self, key: str) -> None:
        if key.startswith('_'):  # pragma: no cover
            return super().__delattr__(key)
        self.delete(key)

    def __delitem__(self, key: str) -> None:
        self.delete(key)

    def delete(self, key: str) -> None:
        """
        Deletes a setting from this object's storage.

        The write to the database is performed immediately and the cache in the cache backend is flushed.
        The cache within this object will be updated correctly.
        """
        if key in self._write_cache():
            self._write_cache()[key].delete()
            del self._write_cache()[key]

        if key in self._cache():
            del self._cache()[key]

        self._flush_external_cache()



from datetime import date, datetime, time
from decimal import Decimal

from django.core.files import File
from django.core.files.storage import default_storage
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase
from django.utils.timezone import now

from hierarkey.models import HierarkeyDefault

from .testapp.models import GlobalSettings, Organization, User, hierarkey


class MyType:
    def __init__(self, foo):
        self.foo = foo

    def __eq__(self, other):
        return self.foo == other.foo


class SettingsTestCase(TestCase):
    def setUp(self):
        hierarkey.add_default('test_default', 'def', str)
        self.global_settings = GlobalSettings()
        self.global_settings.settings.flush()
        self.organization = Organization.objects.create(name='Dummy')
        self.organization.settings.flush()
        self.user = User.objects.create(
            organization=self.organization, name='Dummy'
        )
        self.user.settings.flush()

    def test_global_set_explicit(self):
        self.global_settings.settings.test = 'foo'
        self.assertEqual(self.global_settings.settings.test, 'foo')

        # Reload object
        self.global_settings = GlobalSettings()
        self.assertEqual(self.global_settings.settings.test, 'foo')

    def test_organization_set_explicit(self):
        self.organization.settings.test = 'foo'
        self.assertEqual(self.organization.settings.test, 'foo')

        # Reload object
        self.organization = Organization.objects.get(pk=self.organization.pk)
        self.assertEqual(self.organization.settings.test, 'foo')

    def test_user_set_explicit(self):
        self.user.settings.test = 'foo'
        self.assertEqual(self.user.settings.test, 'foo')

        # Reload object
        self.user = User.objects.get(pk=self.user.pk)
        self.assertEqual(self.user.settings.test, 'foo')

    def test_user_set_twice(self):
        self.user.settings.test = 'bar'
        self.user.settings.test = 'foo'
        self.assertEqual(self.user.settings.test, 'foo')

        # Reload object
        self.user = User.objects.get(pk=self.user.pk)
        self.assertEqual(self.user.settings.test, 'foo')

    def test_organization_set_on_global(self):
        self.global_settings.settings.test = 'foo'
        self.assertEqual(self.global_settings.settings.test, 'foo')
        self.assertEqual(self.organization.settings.test, 'foo')

        # Reload object
        self.global_settings = GlobalSettings()
        self.assertEqual(self.global_settings.settings.test, 'foo')
        self.assertEqual(self.organization.settings.test, 'foo')

    def test_user_set_on_global(self):
        self.global_settings.settings.test = 'foo'
        self.assertEqual(self.global_settings.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'foo')

        # Reload object
        self.global_settings = GlobalSettings()
        self.assertEqual(self.global_settings.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'foo')

    def test_user_set_on_organization(self):
        self.organization.settings.test = 'foo'
        self.assertEqual(self.organization.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'foo')

        # Reload object
        self.organization = Organization.objects.get(id=self.organization.id)
        self.assertEqual(self.organization.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'foo')

    def test_user_override_organization(self):
        self.organization.settings.test = 'foo'
        self.user.settings.test = 'bar'
        self.assertEqual(self.organization.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'bar')

        # Reload object
        self.organization = Organization.objects.get(id=self.organization.id)
        self.user = User.objects.get(id=self.user.id)
        self.assertEqual(self.organization.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'bar')

    def test_user_override_global(self):
        self.global_settings.settings.test = 'foo'
        self.user.settings.test = 'bar'
        self.assertEqual(self.global_settings.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'bar')

        # Reload object
        self.global_settings = GlobalSettings()
        self.user = User.objects.get(id=self.user.id)
        self.assertEqual(self.global_settings.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'bar')

    def test_default(self):
        self.assertEqual(self.global_settings.settings.test_default, 'def')
        self.assertEqual(self.organization.settings.test_default, 'def')
        self.assertEqual(self.user.settings.test_default, 'def')
        self.assertEqual(self.user.settings.get('nonexistant', default='abc'), 'abc')

    def test_default_typing(self):
        self.assertIs(type(self.user.settings.get('nonexistant', as_type=Decimal, default=0)), Decimal)

    def test_item_access(self):
        self.user.settings['foo'] = 'abc'
        self.assertEqual(self.user.settings['foo'], 'abc')
        del self.user.settings['foo']
        self.assertIsNone(self.user.settings['foo'])

    def test_delete(self):
        self.organization.settings.test = 'foo'
        self.user.settings.test = 'bar'
        self.assertEqual(self.organization.settings.test, 'foo')
        self.assertEqual(self.user.settings.test, 'bar')

        del self.user.settings.test
        self.assertEqual(self.user.settings.test, 'foo')

        self.user = User.objects.get(id=self.user.id)
        self.assertEqual(self.user.settings.test, 'foo')

        del self.organization.settings.test
        self.assertIsNone(self.organization.settings.test)

        self.organization = Organization.objects.get(id=self.organization.id)
        self.assertIsNone(self.organization.settings.test)

    def test_serialize_str(self):
        self._test_serialization('ABC', as_type=str)

    def test_serialize_float(self):
        self._test_serialization(2.3, float)

    def test_serialize_int(self):
        self._test_serialization(2, int)

    def test_serialize_datetime(self):
        self._test_serialization(now(), datetime)

    def test_serialize_time(self):
        self._test_serialization(now().time(), time)

    def test_serialize_date(self):
        self._test_serialization(now().date(), date)

    def test_serialize_decimal(self):
        self._test_serialization(Decimal('2.3'), Decimal)

    def test_serialize_dict(self):
        self._test_serialization({'a': 'b', 'c': 'd'}, dict)

    def test_serialize_list(self):
        self._test_serialization([1, 2, 'a'], list)

    def test_serialize_bool(self):
        self._test_serialization(True, bool)
        self._test_serialization(False, bool)

    def test_serialize_bool_implicit(self):
        self.user.settings.set('test', True)
        self.user.settings.flush()
        self.assertIs(self.user.settings.get('test', as_type=None), True)
        self.user.settings.set('test', False)
        self.user.settings.flush()
        self.assertIs(self.user.settings.get('test', as_type=None), False)

    def test_serialize_model(self):
        self._test_serialization(self.user, User)

    def test_serialize_custom_type(self):
        hierarkey.add_type(MyType, lambda v: v.foo, lambda v: MyType(v))
        self._test_serialization(MyType('bar'), MyType)

    def test_custom_type_default(self):
        hierarkey.add_type(MyType, lambda v: v.foo, lambda v: MyType(v))
        hierarkey.add_default('mytype_foo', 'bar', MyType)
        self.assertEqual(self.user.settings.get('mytype_foo'), MyType('bar'))
        self.user.settings.set('mytype_foo', MyType('baz'))
        self.user.settings.flush()
        self.assertEqual(self.user.settings.get('mytype_foo'), MyType('baz'))

    def test_serialize_unknown(self):
        class Type:
            pass

        try:
            self._test_serialization(Type(), Type)
            self.assertTrue(False, 'No exception thrown!')
        except TypeError:
            pass

    def test_serialize_file(self):
        val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
        default_storage.save(val.name, val)
        val.close()
        self.user.settings.set('test', val)
        self.user.settings.flush()
        f = self.user.settings.get('test', as_type=File)
        self.assertIsInstance(f, File)
        self.assertTrue(f.name.endswith(val.name))
        f.close()

    def test_unserialize_file_value(self):
        val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
        default_storage.save(val.name, val)
        val.close()
        self.user.settings.set('test', val)
        self.user.settings.flush()
        f = self.user.settings.get('test', as_type=File)
        self.assertIsInstance(f, File)
        self.assertTrue(f.name.endswith(val.name))
        f.close()

    def test_autodetect_file_value(self):
        val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
        default_storage.save(val.name, val)
        val.close()
        self.user.settings.set('test', val)
        self.user.settings.flush()
        f = self.user.settings.get('test')
        self.assertIsInstance(f, File)
        self.assertTrue(f.name.endswith(val.name))
        f.close()

    def test_autodetect_file_value_of_parent(self):
        val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
        default_storage.save(val.name, val)
        val.close()
        self.organization.settings.set('test', val)
        self.organization.settings.flush()
        f = self.user.settings.get('test')
        self.assertIsInstance(f, File)
        self.assertTrue(f.name.endswith(val.name))
        f.close()

    def test_nonexistant_file(self):
        self.organization.settings.set('test', 'file://foo')
        self.organization.settings.flush()
        f = self.user.settings.get('test', as_type=File)
        self.assertIs(f, False)

    def _test_serialization(self, val, as_type):
        self.user.settings.set('test', val)
        self.user.settings.flush()
        self.assertEqual(self.user.settings.get('test', as_type=as_type), val)
        self.assertIsInstance(self.user.settings.get('test', as_type=as_type), as_type)

    def test_freeze(self):
        olddef = hierarkey.defaults
        hierarkey.defaults = {
            'test_default': HierarkeyDefault(value='def', type=str)
        }
        self.user.organization.settings.set('bar', 'baz')
        self.user.organization.settings.set('foo', 'baz')
        self.user.settings.set('foo', 'bar')
        frozen = self.user.settings.freeze()
        self.user.settings.set('foo', 'notfrozen')

        try:
            self.assertEqual(frozen, {
                'test_default': 'def',
                'bar': 'baz',
                'foo': 'bar'
            })
        finally:
            hierarkey.defaults = olddef



import sys
from collections import namedtuple
from typing import Any, Callable, Optional

from django.core.exceptions import ImproperlyConfigured
from django.db import models


class BaseHierarkeyStoreModel(models.Model):
    key = models.CharField(max_length=255)
    value = models.TextField()

    def __repr__(self):
        return '<{}: "{}"{}>'.format(
            type(self).__name__,
            self.key,
            ' on object {}'.format(self.object_id) if hasattr(self, 'object_id') else '',
        )

    class Meta:
        abstract = True


HierarkeyDefault = namedtuple('HierarkeyDefault', ['value', 'type'])
HierarkeyType = namedtuple('HierarkeyType', ['type', 'serialize', 'unserialize'])


class Hierarkey:
    """
    The Hierarkey object represents one complete key-value store hierarchy. It can have one global and multiple
    object-level storages attached and holds default values as well as custom type serialization info.

    :param attribute_name: The name for the attribute on the model instances that will allow access to the
                           storage, e.g. ``settings``.
    """

    def __init__(self, attribute_name):
        self.attribute_name = attribute_name
        self.global_class = None
        self.defaults = {}
        self.types = []

    def _create_attrs(self, base_model: type) -> dict:
        class Meta:
            pass

        attrs = {
            'Meta': Meta,
            '__module__': base_model.__module__,
        }
        return attrs

    def _create_model(self, model_name: str, attrs: dict) -> type:
        return models.base.ModelBase(model_name, (BaseHierarkeyStoreModel,), attrs)

    def add_default(self, key: str, value: Optional[str], default_type: type = str) -> None:
        """
        Adds a default value and a default type for a key.

        :param key: Key
        :param value: *Serialized* default value, i.e. a string or ``None``.
        :param default_type: The type to deserialize values for this key to, defaults to ``str``.
        """
        self.defaults[key] = HierarkeyDefault(value, default_type)

    def get_declared_type(self, key: str) -> type:
        """
        Returns the type that is declared for a key using add_default.

        :param key: Key
        """
        return self.defaults[key].type if key in self.defaults else None

    def add_type(self, type: type, serialize: Callable[[Any], str], unserialize: Callable[[str], Any]) -> None:
        """
        Adds serialization support for a new type.

        :param type: The type to add support for.
        :param serialize: A callable that takes an object of type ``type`` and returns a string.
        :param unserialize: A callable that takes a string and returns an object of type ``type``.
        """
        self.types.append(HierarkeyType(type=type, serialize=serialize, unserialize=unserialize))

    def set_global(self, cache_namespace: str = None) -> type:
        """
        Decorator. Attaches the global key-value store of this hierarchy to an object.

        :param cache_namespace: Optional. A custom namespace used for caching. By default this is
                                constructed from the name of the class this is applied to and
                                the ``attribute_name`` of this ``Hierarkey`` object.
        """

        if isinstance(cache_namespace, type):
            raise ImproperlyConfigured('Incorrect decorator usage, you need to use .add_global() '
                                       'instead of .add_global')

        def wrapper(wrapped_class):
            if issubclass(wrapped_class, models.Model):
                raise ImproperlyConfigured('Hierarkey.add_global() can only be invoked on a normal class, '
                                           'not on a Django model.')
            if not issubclass(wrapped_class, GlobalSettingsBase):
                raise ImproperlyConfigured('You should use .add_global() on a class that inherits from '
                                           'GlobalSettingsBase.')

            _cache_namespace = cache_namespace or ('%s_%s' % (wrapped_class.__name__, self.attribute_name))

            attrs = self._create_attrs(wrapped_class)

            model_name = '%s_%sStore' % (wrapped_class.__name__, self.attribute_name.title())
            if getattr(sys.modules[wrapped_class.__module__], model_name, None):
                # Already wrapped
                return wrapped_class

            kv_model = self._create_model(model_name, attrs)

            def init(self, *args, object=None, **kwargs):
                super(kv_model, self).__init__(*args, **kwargs)

            setattr(kv_model, '__init__', init)

            hierarkey = self

            def prop(iself):
                from .proxy import HierarkeyProxy

                attrname = '_hierarkey_proxy_{}_{}'.format(_cache_namespace, self.attribute_name)
                cached = getattr(iself, attrname, None)
                if not cached:
                    cached = HierarkeyProxy._new(iself, type=kv_model, hierarkey=hierarkey,
                                                 cache_namespace=_cache_namespace)
                    setattr(iself, attrname, cached)
                return cached

            setattr(sys.modules[wrapped_class.__module__], model_name, kv_model)
            setattr(wrapped_class, '_%s_objects' % self.attribute_name, kv_model.objects)
            setattr(wrapped_class, self.attribute_name, property(prop))
            self.global_class = wrapped_class
            return wrapped_class

        return wrapper

    def add(self, cache_namespace: str = None, parent_field: str = None) -> type:
        """
        Decorator. Attaches a global key-value store to a Django model.

        :param cache_namespace: Optional. A custom namespace used for caching. By default this is
                                constructed from the name of the class this is applied to and
                                the ``attribute_name`` of this ``Hierarkey`` object.
        :param parent_field: Optional. The name of a field of this model that refers to the parent
                             in the hierarchy. This must be a ``ForeignKey`` field.
        """
        if isinstance(cache_namespace, type):
            raise ImproperlyConfigured('Incorrect decorator usage, you need to use .add() instead of .add')

        def wrapper(model):
            if not issubclass(model, models.Model):
                raise ImproperlyConfigured('Hierarkey.add() can only be invoked on a Django model')

            _cache_namespace = cache_namespace or ('%s_%s' % (model.__name__, self.attribute_name))

            attrs = self._create_attrs(model)
            attrs['object'] = models.ForeignKey(model, related_name='_%s_objects' % self.attribute_name,
                                                on_delete=models.CASCADE)
            model_name = '%s_%sStore' % (model.__name__, self.attribute_name.title())
            kv_model = self._create_model(model_name, attrs)

            setattr(sys.modules[model.__module__], model_name, kv_model)

            hierarkey = self

            def prop(iself):
                from .proxy import HierarkeyProxy

                attrname = '_hierarkey_proxy_{}_{}'.format(_cache_namespace, self.attribute_name)
                cached = getattr(iself, attrname, None)
                if not cached:
                    try:
                        parent = getattr(iself, parent_field) if parent_field else None
                    except models.ObjectDoesNotExist:  # pragma: no cover
                        parent = None

                    if not parent and hierarkey.global_class:
                        parent = hierarkey.global_class()

                    cached = HierarkeyProxy._new(
                        iself,
                        type=kv_model,
                        hierarkey=hierarkey,
                        parent=parent,
                        cache_namespace=_cache_namespace
                    )
                    setattr(iself, attrname, cached)
                return cached

            setattr(model, self.attribute_name, property(prop))

            return model

        return wrapper


class GlobalSettingsBase:
    """
    Base class for objects with a global settings storage attached. This class does not
    add any functionality, it only makes global settings behave more consistent to
    object-level settings.
    """

    pk = '_global'



import logging

from django import forms
from django.core.exceptions import FieldDoesNotExist
from django.core.files import File
from django.core.files.storage import default_storage
from django.core.files.uploadedfile import UploadedFile
from django.utils.crypto import get_random_string
from django.utils.text import normalize_newlines
from django.utils.translation import gettext_lazy as _

from hierarkey.models import BaseHierarkeyStoreModel

logger = logging.getLogger(__name__)


class HierarkeyForm(forms.Form):
    """
    This is a custom subclass of ``django.forms.Form`` that you can use to set
    values for any keys. See the Forms chapter of the documentation for more details.
    """
    BOOL_CHOICES = (
        ('False', _('disabled')),
        ('True', _('enabled')),
    )

    def __init__(self, *args, obj, attribute_name, **kwargs):
        self.obj = obj
        self.attribute_name = attribute_name
        self._s = getattr(obj, attribute_name)
        initial = kwargs.pop('initial', {})
        initial.update(self._s.freeze())
        kwargs['initial'] = initial
        super().__init__(*args, **kwargs)

    def save(self) -> None:
        """
        Saves all changed values to the database.
        """
        for name, field in self.fields.items():
            value = self.cleaned_data[name]
            if isinstance(value, UploadedFile):
                # Delete old file
                fname = self._s.get(name, as_type=File)
                if fname and self.__file_is_last_reference(name, self._s.get(name, as_type=str)):
                    try:
                        default_storage.delete(fname.name)
                    except OSError:  # pragma: no cover
                        logger.error('Deleting file %s failed.' % fname.name)

                # Create new file
                newname = default_storage.save(self.get_new_filename(value.name), value)
                value._name = newname
                self._s.set(name, value)
            elif isinstance(value, File):
                # file is unchanged
                continue
            elif not value and isinstance(field, forms.FileField):
                # file is deleted
                fname = self._s.get(name, as_type=File)
                if fname and self.__file_is_last_reference(name, self._s.get(name, as_type=str)):
                    try:
                        default_storage.delete(fname.name)
                    except OSError:  # pragma: no cover
                        logger.error('Deleting file %s failed.' % fname.name)
                del self._s[name]
            elif value is None:
                del self._s[name]
            elif type(value) is str:
                if normalize_newlines(self._s.get(name, as_type=str)) != normalize_newlines(value):
                    self._s.set(name, value)
            elif self._s.get(name, as_type=type(value)) != value:
                self._s.set(name, value)

    def __file_is_last_reference(self, key, value):
        for klass in BaseHierarkeyStoreModel.__subclasses__():
            if klass._meta.abstract:
                continue
            qs = klass.objects.filter(key=key, value=value)
            try:
                if isinstance(self.obj, klass._meta.get_field('object').remote_field.model):
                    qs = qs.exclude(object=self.obj)
            except FieldDoesNotExist:
                pass
            print(klass, qs)
            if qs.exists():
                print("false", qs, self.obj, klass)
                return False
        return True

    def get_new_filename(self, name: str) -> str:
        """
        Returns the file name to use based on the original filename of an uploaded file.
        By default, the file name is constructed as::

            <model_name>-<attribute_name>/<primary_key>/<original_basename>.<random_nonce>.<extension>
        """
        nonce = get_random_string(length=8)
        return '%s-%s/%s/%s.%s.%s' % (
            self.obj._meta.model_name, self.attribute_name,
            self.obj.pk, name, nonce, name.split('.')[-1]
        )



import os

import pytest
from django import forms
from django.core.files import File
from django.core.files.uploadedfile import SimpleUploadedFile

from hierarkey.forms import HierarkeyForm

from .testapp.models import Organization, hierarkey


class SampleForm(HierarkeyForm):
    test_string = forms.CharField(max_length=100, required=False)
    test_int = forms.IntegerField(required=False)
    test_file = forms.FileField(required=False)


@pytest.fixture
def organization():
    o = Organization.objects.create(name='Foo')
    o.settings.flush()
    return o


@pytest.mark.django_db
def test_form_defaults(organization):
    hierarkey.add_default('test_string', 'hallo')
    form = SampleForm(obj=organization, attribute_name='settings')
    assert form.initial['test_string'] == 'hallo'


@pytest.mark.django_db
def test_form_initial_values(organization):
    hierarkey.add_default('test_string', 'hallo')
    organization.settings.test_string = 'bar'
    form = SampleForm(obj=organization, attribute_name='settings')
    assert form.initial['test_string'] == 'bar'


@pytest.mark.django_db
def test_form_save_plain_values(organization):
    form = SampleForm(obj=organization, attribute_name='settings', data={
        'test_string': 'foobar',
        'test_int': 42
    })
    assert form.is_valid()
    form.save()
    organization.settings.flush()
    assert organization.settings.test_string == 'foobar'
    assert organization.settings.test_int == '42'


@pytest.mark.django_db
def test_form_ignore_newline_change(organization):
    form = SampleForm(obj=organization, attribute_name='settings', data={
        'test_string': 'foo\nbar',
        'test_int': 42
    })
    assert form.is_valid()
    form.save()
    organization.settings.flush()
    assert organization.settings.test_string == 'foo\nbar'
    assert organization.settings.test_int == '42'

    form = SampleForm(obj=organization, attribute_name='settings', data={
        'test_string': 'foo\r\nbar',
        'test_int': 42
    })
    assert form.is_valid()
    form.save()
    organization.settings.flush()
    assert organization.settings.test_string == 'foo\nbar'
    assert organization.settings.test_int == '42'


@pytest.mark.django_db
def test_form_save_new_file(organization):
    val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
    form = SampleForm(obj=organization, attribute_name='settings', data={}, files={
        'test_file': val
    })
    assert form.is_valid()
    form.save()
    organization.settings.flush()
    assert organization.settings.get('test_file', as_type=str).startswith('file://')
    assert organization.settings.get('test_file', as_type=str).endswith(val.name)
    assert organization.settings.get('test_file', as_type=str).endswith(val.name)
    assert os.path.exists(organization.settings.get('test_file', as_type=File, binary_file=True).name)


@pytest.mark.django_db
def test_form_save_replaced_file(organization):
    val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
    form = SampleForm(obj=organization, attribute_name='settings', data={}, files={
        'test_file': val
    })
    assert form.is_valid()
    form.save()

    organization.settings.flush()
    oldname = organization.settings.get('test_file', as_type=File, binary_file=True).name

    val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content_2", content_type="image/jpeg")
    form = SampleForm(obj=organization, attribute_name='settings', data={}, files={
        'test_file': val
    })
    assert form.is_valid()
    form.save()

    organization.settings.flush()
    assert os.path.exists(organization.settings.get('test_file', as_type=File, binary_file=True).name)
    assert organization.settings.get('test_file', as_type=File, binary_file=True).name != oldname
    assert not os.path.exists(oldname)


@pytest.mark.django_db
def test_form_save_unchanged_file(organization):
    val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
    form = SampleForm(obj=organization, attribute_name='settings', data={}, files={
        'test_file': val
    })
    assert form.is_valid()
    form.save()

    organization.settings.flush()
    oldname = organization.settings.get('test_file', as_type=File, binary_file=True).name

    form = SampleForm(obj=organization, attribute_name='settings', data={'unaffected': 'on'}, files={})
    assert form.is_valid()
    form.save()

    organization.settings.flush()
    assert organization.settings.get('test_file', as_type=File, binary_file=True).name == oldname


@pytest.mark.django_db
def test_form_delete_file(organization):
    val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
    form = SampleForm(obj=organization, attribute_name='settings', data={}, files={
        'test_file': val
    })
    assert form.is_valid()
    form.save()

    organization.settings.flush()
    oldname = organization.settings.get('test_file', as_type=File, binary_file=True).name

    form = SampleForm(obj=organization, attribute_name='settings', data={
        'test_file-clear': 'on'
    })
    assert form.is_valid()
    form.save()

    organization.settings.flush()
    assert not organization.settings.test_file
    assert not os.path.exists(oldname)


@pytest.mark.django_db
def test_form_do_not_delete_file_if_referenced(organization):
    val = SimpleUploadedFile("sample_invalid_image.jpg", b"file_content", content_type="image/jpeg")
    form = SampleForm(obj=organization, attribute_name='settings', data={}, files={
        'test_file': val
    })
    assert form.is_valid()
    form.save()

    organization.settings.flush()

    org2 = Organization.objects.create(name='Bar')
    org2.settings.test_file = organization.settings.get('test_file', as_type=str)

    oldname = organization.settings.get('test_file', as_type=File, binary_file=True).name

    form = SampleForm(obj=organization, attribute_name='settings', data={
        'test_file-clear': 'on'
    })
    assert form.is_valid()
    form.save()

    organization.settings.flush()
    assert not organization.settings.test_file
    assert os.path.exists(oldname)

    form = SampleForm(obj=org2, attribute_name='settings', data={
        'test_file-clear': 'on'
    })
    assert form.is_valid()
    form.save()

    organization.settings.flush()
    assert not org2.settings.test_file
    assert not os.path.exists(oldname)



import pytest
from django.core.exceptions import ImproperlyConfigured

from .testapp.models import GlobalSettings, User, hierarkey


def test_global_settings_on_invalid():
    with pytest.raises(ImproperlyConfigured):
        hierarkey.set_global()(object)


def test_global_settings_on_model():
    with pytest.raises(ImproperlyConfigured):
        hierarkey.set_global()(User)


def test_model_settings_on_non_model():
    with pytest.raises(ImproperlyConfigured):
        hierarkey.add()(GlobalSettings)


def test_missing_parantheses():
    with pytest.raises(ImproperlyConfigured):
        hierarkey.set_global(GlobalSettings)

    with pytest.raises(ImproperlyConfigured):
        hierarkey.add(User)
